# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _partition module.
"""


import re
import copy
import logging
import pytest

from zhmcclient import Client, Partition, HTTPError, NotFound
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources, assert_blanked_in_message

# Object IDs and names of our faked partitions:
PART1_OID = 'part1-oid'
PART1_NAME = 'part 1'
PART2_OID = 'part2-oid'
PART2_NAME = 'part 2'
PART3_OID = 'part3-oid'
PART3_NAME = 'part 3'

CPC_NAME = 'fake-cpc1-name'

# Properties returned by default from list_permitted_partitions()
LIST_PERMITTED_PARTITIONS_PROPS = [
    'name', 'object-uri', 'type', 'status', 'has-unacceptable-status',
    'cpc-name', 'cpc-object-uri',
    # The zhmcclient_mock support always returns 'se-version'
    'se-version'
]


class TestPartition:
    """All tests for the Partition and PartitionManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked CPC in DPM mode without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_cpc = self.session.hmc.cpcs.add({
            'object-id': 'fake-cpc1-oid',
            # object-uri is set up automatically
            'parent': None,
            'class': 'cpc',
            'name': CPC_NAME,
            'description': 'CPC #1 (DPM mode)',
            'status': 'active',
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'iml-mode': 'dpm',
        })
        self.cpc = self.client.cpcs.find(name=CPC_NAME)

    def add_partition1(self):
        """Add partition 1 (type linux)."""

        faked_partition = self.faked_cpc.partitions.add({
            'object-id': PART1_OID,
            # object-uri will be automatically set
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': PART1_NAME,
            'description': 'Partition #1',
            'status': 'active',
            'type': 'linux',
            'ifl-processors': 2,
            'initial-memory': 4096,
            'maximum-memory': 8192,
        })
        return faked_partition

    def add_partition2(self):
        """Add partition 2 (type ssc)."""

        faked_partition = self.faked_cpc.partitions.add({
            'object-id': PART2_OID,
            # object-uri will be automatically set
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': PART2_NAME,
            'description': 'Partition #2',
            'status': 'active',
            'type': 'ssc',
            'ifl-processors': 2,
            'initial-memory': 4096,
            'maximum-memory': 8192,
        })
        return faked_partition

    def add_partition3(self):
        """Add partition 3 (support for firmware features)."""

        faked_partition = self.faked_cpc.partitions.add({
            'object-id': PART3_OID,
            # object-uri will be automatically set
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': PART3_NAME,
            'description': 'Partition #3',
            'status': 'active',
            'type': 'linux',
            'ifl-processors': 2,
            'initial-memory': 4096,
            'maximum-memory': 8192,
            'available-features-list': [],
        })
        return faked_partition

    def add_partition(self, part_name):
        """Add a partition (using one of the known names)."""

        if part_name == PART1_NAME:
            faked_partition = self.add_partition1()
        elif part_name == PART2_NAME:
            faked_partition = self.add_partition2()
        else:
            assert part_name == PART3_NAME
            faked_partition = self.add_partition3()
        return faked_partition

    def test_pm_initial_attrs(self):
        """Test initial attributes of PartitionManager."""

        partition_mgr = self.cpc.partitions

        # Verify all public properties of the manager object
        assert partition_mgr.resource_class == Partition
        assert partition_mgr.session == self.session
        assert partition_mgr.parent == self.cpc
        assert partition_mgr.cpc == self.cpc

    # TODO: Test for PartitionManager.__repr__()

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            ({},
             ['object-uri', 'name', 'status']),
            (dict(full_properties=False),
             ['object-uri', 'name', 'status']),
            (dict(full_properties=True),
             None),
        ]
    )
    def test_pm_list_full_properties(
            self, full_properties_kwargs, prop_names):
        """Test PartitionManager.list() with full_properties."""

        # Add two faked partitions
        faked_partition1 = self.add_partition1()
        faked_partition2 = self.add_partition2()

        exp_faked_partitions = [faked_partition1, faked_partition2]
        partition_mgr = self.cpc.partitions

        # Execute the code to be tested
        partitions = partition_mgr.list(**full_properties_kwargs)

        assert_resources(partitions, exp_faked_partitions, prop_names)

    @pytest.mark.parametrize(
        "filter_args, exp_names", [
            ({'object-id': PART1_OID},
             [PART1_NAME]),
            ({'object-id': PART2_OID},
             [PART2_NAME]),
            ({'object-id': [PART1_OID, PART2_OID]},
             [PART1_NAME, PART2_NAME]),
            ({'object-id': [PART1_OID, PART1_OID]},
             [PART1_NAME]),
            ({'object-id': PART1_OID + 'foo'},
             []),
            ({'object-id': [PART1_OID, PART2_OID + 'foo']},
             [PART1_NAME]),
            ({'object-id': [PART2_OID + 'foo', PART1_OID]},
             [PART1_NAME]),
            ({'name': PART1_NAME},
             [PART1_NAME]),
            ({'name': PART2_NAME},
             [PART2_NAME]),
            ({'name': [PART1_NAME, PART2_NAME]},
             [PART1_NAME, PART2_NAME]),
            ({'name': PART1_NAME + 'foo'},
             []),
            ({'name': [PART1_NAME, PART2_NAME + 'foo']},
             [PART1_NAME]),
            ({'name': [PART2_NAME + 'foo', PART1_NAME]},
             [PART1_NAME]),
            ({'name': [PART1_NAME, PART1_NAME]},
             [PART1_NAME]),
            ({'name': '.*part 1'},
             [PART1_NAME]),
            ({'name': 'part 1.*'},
             [PART1_NAME]),
            ({'name': 'part .'},
             [PART1_NAME, PART2_NAME]),
            ({'name': '.art 1'},
             [PART1_NAME]),
            ({'name': '.+'},
             [PART1_NAME, PART2_NAME]),
            ({'name': 'part 1.+'},
             []),
            ({'name': '.+part 1'},
             []),
            ({'name': PART1_NAME,
              'object-id': PART1_OID},
             [PART1_NAME]),
            ({'name': PART1_NAME,
              'object-id': PART1_OID + 'foo'},
             []),
            ({'name': PART1_NAME + 'foo',
              'object-id': PART1_OID},
             []),
            ({'name': PART1_NAME + 'foo',
              'object-id': PART1_OID + 'foo'},
             []),
        ]
    )
    def test_pm_list_filter_args(self, filter_args, exp_names):
        """Test PartitionManager.list() with filter_args."""

        # Add two faked partitions
        self.add_partition1()
        self.add_partition2()

        partition_mgr = self.cpc.partitions

        # Execute the code to be tested
        partitions = partition_mgr.list(filter_args=filter_args)

        assert len(partitions) == len(exp_names)
        if exp_names:
            names = [p.properties['name'] for p in partitions]
            assert set(names) == set(exp_names)

    @pytest.mark.parametrize(
        "list_kwargs, prop_names", [
            ({},
             ['object-uri', 'name', 'status']),
            (dict(additional_properties=[]),
             ['object-uri', 'name', 'status']),
            (dict(additional_properties=['description']),
             ['object-uri', 'name', 'status', 'description']),
            (dict(additional_properties=['description', 'se-version']),
             ['object-uri', 'name', 'status', 'description', 'se-version']),
            (dict(additional_properties=['ssc-host-name']),
             ['object-uri', 'name', 'status', 'ssc-host-name']
             # ssc-host-name is not on every partition
             ),
        ]
    )
    def test_pm_list_add_props(
            self, list_kwargs, prop_names):
        """
        Test PartitionManager.list() with additional_properties.
        """

        # Add two faked partitions
        faked_part1 = self.add_partition1()
        faked_part2 = self.add_partition2()

        partition_mgr = self.cpc.partitions

        # Execute the code to be tested
        parts = partition_mgr.list(**list_kwargs)

        exp_faked_parts = [faked_part1, faked_part2]

        assert_resources(parts, exp_faked_parts, prop_names)

    @pytest.mark.parametrize(
        "input_props, exp_prop_names, exp_exc", [
            ({},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'description': 'fake description X'},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-part-x'},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-part-x',
              'initial-memory': 1024},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-part-x',
              'ifl-processors': 2,
              'initial-memory': 4096,
              'maximum-memory': 4096,
              'description': 'fake description X'},
             ['object-uri', 'name', 'initial-memory', 'maximum-memory',
              'ifl-processors', 'description'],
             None),
            ({'name': 'fake-part-x',
              'ifl-processors': 2,
              'initial-memory': 4096,
              'maximum-memory': 4096,
              'boot-device': 'ftp',
              'boot-ftp-host': 'host',
              'boot-ftp-username': 'user',
              'boot-ftp-password': 'bla',
              'boot-ftp-insfile': 'ins'},
             ['object-uri', 'name', 'initial-memory', 'maximum-memory',
              'ifl-processors', 'boot-device', 'boot-ftp-host',
              'boot-ftp-username', 'boot-ftp-insfile'],
             None),
            ({'name': 'fake-part-x',
              'ifl-processors': 2,
              'initial-memory': 4096,
              'maximum-memory': 4096,
              'type': 'ssc',
              'ssc-host-name': 'host',
              'ssc-master-userid': 'user',
              'ssc-master-pw': 'bla'},
             ['object-uri', 'name', 'initial-memory', 'maximum-memory',
              'ifl-processors', 'type', 'ssc-host-name', 'ssc-master-userid'],
             None),
        ]
    )
    def test_pm_create(self, caplog, input_props, exp_prop_names, exp_exc):
        """Test PartitionManager.create()."""

        logger_name = "zhmcclient.api"
        caplog.set_level(logging.DEBUG, logger=logger_name)

        partition_mgr = self.cpc.partitions

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                partition = partition_mgr.create(properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            # Note: the Partition object returned by Partition.create() has
            # the input properties plus 'object-uri'.
            partition = partition_mgr.create(properties=input_props)

            # Get its API call log record
            call_record = caplog.records[-2]

            # Check the resource for consistency within itself
            assert isinstance(partition, Partition)
            partition_name = partition.name
            exp_partition_name = partition.properties['name']
            assert partition_name == exp_partition_name
            partition_uri = partition.uri
            exp_partition_uri = partition.properties['object-uri']
            assert partition_uri == exp_partition_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in partition.properties
                if prop_name in input_props:
                    value = partition.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

            # Verify the API call log record for blanked-out properties.
            assert_blanked_in_message(
                call_record.message, input_props,
                ['boot-ftp-password', 'ssc-master-pw'])

    def test_pm_resource_object(self):
        """
        Test PartitionManager.resource_object().

        This test exists for historical reasons, and by now is covered by the
        test for BaseManager.resource_object().
        """
        partition_mgr = self.cpc.partitions

        partition_oid = 'fake-partition-id42'

        # Execute the code to be tested
        partition = partition_mgr.resource_object(partition_oid)

        partition_uri = "/api/partitions/" + partition_oid

        assert isinstance(partition, Partition)
        assert partition.uri == partition_uri
        assert partition.properties['object-uri'] == partition_uri
        assert partition.properties['object-id'] == partition_oid
        assert partition.properties['class'] == 'partition'
        assert partition.properties['parent'] == self.cpc.uri

    # TODO: Test for initial Partition attributes (nics, hbas,
    #       virtual_functions)

    def test_partition_repr(self):
        """Test Partition.__repr__()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        # Execute the code to be tested
        repr_str = repr(partition)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{partition.__class__.__name__}\s+at\s+'
            rf'0x{id(partition):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "initial_status, exp_exc", [
            ('stopped', None),
            ('terminated', HTTPError({'http-status': 409, 'reason': 1})),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', HTTPError({'http-status': 409, 'reason': 1})),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', HTTPError({'http-status': 409, 'reason': 1})),
            ('reservation-error',
             HTTPError({'http-status': 409, 'reason': 1})),
            ('paused', HTTPError({'http-status': 409, 'reason': 1})),
        ]
    )
    def test_partition_delete(self, initial_status, exp_exc):
        """Test Partition.delete()."""

        # Add a faked partition to be tested and another one
        faked_partition = self.add_partition1()
        self.add_partition2()

        # Set the initial status of the faked partition
        faked_partition.properties['status'] = initial_status

        partition_mgr = self.cpc.partitions

        partition = partition_mgr.find(name=faked_partition.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                partition.delete()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the partition still exists
            partition_mgr.find(name=faked_partition.name)

        else:

            # Execute the code to be tested.
            partition.delete()

            # Check that the partition no longer exists
            with pytest.raises(NotFound) as exc_info:
                partition_mgr.find(name=faked_partition.name)

    def test_partition_delete_create_same_name(self):
        """Test Partition.delete() followed by create() with same name."""

        # Add a faked partition to be tested and another one
        faked_partition = self.add_partition1()
        partition_name = faked_partition.name
        self.add_partition2()

        # Construct the input properties for a third partition
        part3_props = {
            'name': partition_name,
            'description': 'Third partition',
            'ifl-processors': 4,
            'initial-memory': 4096,
            'maximum-memory': 8192,
        }

        # Set the initial status of the faked partition
        faked_partition.properties['status'] = 'stopped'  # deletable

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=partition_name)

        # Execute the deletion code to be tested.
        partition.delete()

        # Check that the partition no longer exists
        with pytest.raises(NotFound):
            partition_mgr.find(name=partition_name)

        # Execute the creation code to be tested.
        partition_mgr.create(part3_props)

        # Check that the partition exists again under that name
        partition3 = partition_mgr.find(name=partition_name)
        description = partition3.get_property('description')
        assert description == 'Third partition'

    @pytest.mark.parametrize(
        "desc, partition_name, available_features, feature_name, "
        "exp_feature_enabled, exp_exc", [
            (
                "No feature support on the CPC",
                PART1_NAME,
                None,
                'fake-feature1', None, ValueError()
            ),
            (
                "Feature not available on the partition (empty feature list)",
                PART3_NAME,
                [],
                'fake-feature1', None, ValueError()
            ),
            (
                "Feature not available on the part (one other feature avail)",
                PART3_NAME,
                [
                    dict(name='fake-feature-foo', state=True),
                ],
                'fake-feature1', None, ValueError()
            ),
            (
                "Feature disabled (the only feature available)",
                PART3_NAME,
                [
                    dict(name='fake-feature1', state=False),
                ],
                'fake-feature1', False, None
            ),
            (
                "Feature enabled (the only feature available)",
                PART3_NAME,
                [
                    dict(name='fake-feature1', state=True),
                ],
                'fake-feature1', True, None
            ),
        ]
    )
    def test_partition_feature_enabled(
            self, desc, partition_name, available_features, feature_name,
            exp_feature_enabled, exp_exc):
        # pylint: disable=unused-argument
        """Test Partition.feature_enabled()."""

        # Add a faked Partition
        faked_partition = self.add_partition(partition_name)

        # Set up the firmware feature list
        if available_features is not None:
            faked_partition.properties['available-features-list'] = \
                available_features

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=partition_name)

        if exp_exc:
            with pytest.raises(exp_exc.__class__):

                # Execute the code to be tested
                partition.feature_enabled(feature_name)

        else:

            # Execute the code to be tested
            act_feature_enabled = partition.feature_enabled(feature_name)

            assert act_feature_enabled == exp_feature_enabled

    @pytest.mark.parametrize(
        "desc, partition_name, available_features, exp_exc", [
            (
                "No feature support on the CPC",
                PART1_NAME,
                None,
                ValueError()
            ),
            (
                "Feature not available on the partition (empty feature list)",
                PART3_NAME,
                [],
                None
            ),
            (
                "Feature not available on the part (one other feature avail)",
                PART3_NAME,
                [
                    dict(name='fake-feature-foo', state=True),
                ],
                None
            ),
            (
                "Feature disabled (the only feature available)",
                PART3_NAME,
                [
                    dict(name='fake-feature1', state=False),
                ],
                None
            ),
            (
                "Feature enabled (the only feature available)",
                PART3_NAME,
                [
                    dict(name='fake-feature1', state=True),
                ],
                None
            ),
        ]
    )
    def test_partition_feature_info(
            self, desc, partition_name, available_features, exp_exc):
        # pylint: disable=unused-argument
        """Test Partition.feature_info()."""

        # Add a faked Partition
        faked_partition = self.add_partition(partition_name)

        # Set up the firmware feature list
        if available_features is not None:
            faked_partition.properties['available-features-list'] = \
                available_features

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=partition_name)

        if exp_exc:
            with pytest.raises(exp_exc.__class__):

                # Execute the code to be tested
                partition.feature_info()

        else:

            # Execute the code to be tested
            act_features = partition.feature_info()

            assert act_features == available_features

    @pytest.mark.parametrize(
        "partition_name", [
            PART1_NAME,
            PART2_NAME,
        ]
    )
    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New partition description'},
            {'initial-memory': 512,
             'description': 'New partition description'},
            {'autogenerate-partition-id': True,
             'partition-id': None},
            {'boot-device': 'none',
             'boot-ftp-host': None,
             'boot-ftp-username': None,
             'boot-ftp-password': None,
             'boot-ftp-insfile': None},
            {'boot-device': 'none',
             'boot-network-device': None},
            {'boot-device': 'none',
             'boot-removable-media': None,
             'boot-removable-media-type': None},
            {'boot-device': 'none',
             'boot-storage-device': None,
             'boot-logical-unit-number': None,
             'boot-world-wide-port-name': None},
            {'boot-device': 'none',
             'boot-iso-image-name': None,
             'boot-iso-insfile': None},
            {'ssc-ipv4-gateway': None,
             'ssc-ipv6-gateway': None,
             'ssc-master-userid': None,
             'ssc-master-pw': None},
        ]
    )
    def test_partition_update_properties(
            self, caplog, input_props, partition_name):
        """Test Partition.update_properties()."""

        logger_name = "zhmcclient.api"
        caplog.set_level(logging.DEBUG, logger=logger_name)

        # Add faked partitions
        self.add_partition1()
        self.add_partition2()

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=partition_name)

        partition.pull_full_properties()
        saved_properties = copy.deepcopy(partition.properties)

        # Execute the code to be tested
        partition.update_properties(properties=input_props)

        # Get its API call log record
        call_record = caplog.records[-2]

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in partition.properties
            prop_value = partition.properties[prop_name]
            assert prop_value == exp_prop_value

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        partition.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in partition.properties
            prop_value = partition.properties[prop_name]
            assert prop_value == exp_prop_value

        # Verify the API call log record for blanked-out properties.
        assert_blanked_in_message(
            call_record.message, input_props,
            ['boot-ftp-password', 'ssc-master-pw'])

    def test_partition_update_name(self):
        """
        Test Partition.update_properties() with 'name' property.
        """

        # Add a faked partition
        faked_partition = self.add_partition1()
        partition_name = faked_partition.name

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=partition_name)

        new_partition_name = "new-" + partition_name

        # Execute the code to be tested
        partition.update_properties(properties={'name': new_partition_name})

        # Verify that the resource is no longer found by its old name, using
        # list() (this does not use the name-to-URI cache).
        partitions_list = partition_mgr.list(
            filter_args=dict(name=partition_name))
        assert len(partitions_list) == 0

        # Verify that the resource is no longer found by its old name, using
        # find() (this uses the name-to-URI cache).
        with pytest.raises(NotFound):
            partition_mgr.find(name=partition_name)

        # Verify that the resource object already reflects the update, even
        # though it has not been refreshed yet.
        assert partition.properties['name'] == new_partition_name

        # Refresh the resource object and verify that it still reflects the
        # update.
        partition.pull_full_properties()
        assert partition.properties['name'] == new_partition_name

        # Verify that the resource can be found by its new name, using find()
        new_partition_find = partition_mgr.find(name=new_partition_name)
        assert new_partition_find.properties['name'] == new_partition_name

        # Verify that the resource can be found by its new name, using list()
        new_partitions_list = partition_mgr.list(
            filter_args=dict(name=new_partition_name))
        assert len(new_partitions_list) == 1
        new_partition_list = new_partitions_list[0]
        assert new_partition_list.properties['name'] == new_partition_name

    @pytest.mark.parametrize(
        "initial_status, exp_exc", [
            ('stopped', None),
            ('terminated', HTTPError({'http-status': 409, 'reason': 1})),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', HTTPError({'http-status': 409, 'reason': 1})),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', HTTPError({'http-status': 409, 'reason': 1})),
            ('reservation-error',
             HTTPError({'http-status': 409, 'reason': 1})),
            ('paused', HTTPError({'http-status': 409, 'reason': 1})),
        ]
    )
    def test_partition_start(self, initial_status, exp_exc):
        """Test Partition.start()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        # Set the initial status of the faked partition
        faked_partition.properties['status'] = initial_status

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                partition.start()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            ret = partition.start()

            assert ret == {}

            partition.pull_full_properties()
            status = partition.get_property('status')
            assert status == 'active'

    @pytest.mark.parametrize(
        "initial_status, exp_exc", [
            ('stopped', HTTPError({'http-status': 409, 'reason': 1})),
            ('terminated', None),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', None),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', HTTPError({'http-status': 409, 'reason': 1})),
            ('reservation-error',
             HTTPError({'http-status': 409, 'reason': 1})),
            ('paused', None),
        ]
    )
    def test_partition_stop(self, initial_status, exp_exc):
        """Test Partition.stop()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        # Set the initial status of the faked partition
        faked_partition.properties['status'] = initial_status

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                partition.stop()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            ret = partition.stop()

            assert ret == {}

            partition.pull_full_properties()
            status = partition.get_property('status')
            assert status == 'stopped'

    # TODO: Re-enable test_partition_dump_partition() once supported in hdlr
    def xtest_partition_dump_partition(self):
        """Test Partition.dump_partition()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        parameters = {
            'dump-load-hba-uri': 'fake-hba-uri',
            'dump-world-wide-port-name': 'fake-wwpn',
            'dump-logical-unit-number': 'fake-lun',
        }

        # Execute the code to be tested.
        ret = partition.dump_partition(parameters=parameters)

        assert ret == {}

    def test_partition_start_dump_program(self):
        """Test Partition.start_dump_program()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        parameters = {
            'dump-program-type': 'storage',
            'dump-program-info': {
                'storage-volume-uri': 'sv1'  # dummy
            },
        }

        # Execute the code to be tested.
        ret = partition.start_dump_program(parameters=parameters)

        assert ret == {}

    # TODO: Re-enable test_partition_psw_restart() once supported in hdlr
    def xtest_partition_psw_restart(self):
        """Test Partition.psw_restart()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        # Execute the code to be tested.
        ret = partition.psw_restart()

        assert ret == {}

    # TODO: Re-enable test_partition_mount_iso_image() once supported in hdlr
    def xtest_partition_mount_iso_image(self):
        """Test Partition.mount_iso_image()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        # TODO: Add test case where a file-like object is passed as image
        image = b'fake-image-data'
        image_name = 'fake-image-name'
        ins_file_name = 'fake-ins-file-name'

        # Execute the code to be tested.
        ret = partition.mount_iso_image(image=image, image_name=image_name,
                                        ins_file_name=ins_file_name)

        assert ret is None

    # TODO: Re-enable test_partition_unmount_iso_image() once supported in hdlr
    def xtest_partition_unmount_iso_image(self):
        """Test Partition.unmount_iso_image()."""

        # Add a faked partition
        faked_partition = self.add_partition1()

        partition_mgr = self.cpc.partitions
        partition = partition_mgr.find(name=faked_partition.name)

        # Execute the code to be tested.
        ret = partition.unmount_iso_image()

        assert ret is None

    @pytest.mark.parametrize(
        "filter_args, additional_props, exp_names", [
            ({'cpc-name': 'bad'},
             None,
             []),
            ({'cpc-name': CPC_NAME},
             ['ifl-processors', 'maximum-memory'],
             [PART1_NAME, PART2_NAME]),
            ({},
             None,
             [PART1_NAME, PART2_NAME]),
            (None,
             None,
             [PART1_NAME, PART2_NAME]),
            ({'name': PART1_NAME},
             ['maximum-memory'],
             [PART1_NAME]),
            ({'name': PART1_NAME, 'cpc-name': CPC_NAME},
             None,
             [PART1_NAME]),
            ({'name': PART1_NAME, 'cpc-name': 'bad'},
             None,
             [])
        ]
    )
    def test_console_list_permitted_partitions(self, filter_args,
                                               additional_props, exp_names):
        """Test Console.list_permitted_partitions() with filter_args."""

        # Add two faked partitions
        self.add_partition1()
        self.add_partition2()

        self.session.hmc.consoles.add({
            'object-id': None,
            # object-uri will be automatically set
            'parent': None,
            'class': 'console',
            'name': 'fake-console1',
            'description': 'Console #1',
        })

        console = self.client.consoles.console

        # Execute the code to be tested
        partitions = console.list_permitted_partitions(
            filter_args=filter_args,
            additional_properties=additional_props)

        assert len(partitions) == len(exp_names)
        if exp_names:
            names = [p.properties['name'] for p in partitions]
            assert set(names) == set(exp_names)

        for partition in partitions:
            partition_props = dict(partition.properties)
            for pname in LIST_PERMITTED_PARTITIONS_PROPS:
                assert pname in partition_props, (
                    f"Property {pname!r} missing from returned partition "
                    f"properties, got: {partition_props!r}")
            if additional_props:
                for pname in additional_props:
                    assert pname in partition_props, (
                        f"Property {pname!r} missing from returned partition "
                        f"properties, got: {partition_props!r}")

    # TODO: Test for Partition.send_os_command()

    # TODO: Test for Partition.create_os_websocket()

    # TODO: Test for Partition.wait_for_status()

    # TODO: Test for Partition.increase_crypto_config()

    # TODO: Test for Partition.decrease_crypto_config()

    # TODO: Test for Partition.change_crypto_domain_config()

    # TODO: Test for Partition.zeroize_crypto_domain()

    # TODO: Test for Partition.attach_storage_group()

    # TODO: Test for Partition.detach_storage_group()

    # TODO: Test for Partition.list_attached_storage_groups()



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for tasks (on CPCs in DPM mode).

These tests do not change any existing tasks.
"""


from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, runtest_find_list, \
    runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic Task objects (e.g. find_by_name())
TASK_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in Task objects returned by list() without full props
TASK_LIST_PROPS = ['element-uri', 'name']

# Properties whose values can change between retrievals of Task objects
TASK_VOLATILE_PROPS = []


def test_task_find_list(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "tasks")

    # Pick the tasks to test with
    task_list = console.tasks.list()
    assert task_list  # system-defined and therefore never empty
    task_list = pick_test_resources(task_list)

    for task in task_list:
        print(f"Testing with task {task.name!r}")
        runtest_find_list(
            hmc_session, console.tasks, task.name, 'name', 'element-uri',
            TASK_VOLATILE_PROPS, TASK_MINIMAL_PROPS, TASK_LIST_PROPS)


def test_task_property(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "tasks")

    # Pick the tasks to test with
    task_list = console.tasks.list()
    assert task_list  # system-defined and therefore never empty
    task_list = pick_test_resources(task_list)

    for task in task_list:
        print(f"Testing with task {task.name!r}")

        # Select a property that is not returned by list()
        non_list_prop = 'description'

        runtest_get_properties(task.manager, non_list_prop)



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A faked Session class for the zhmcclient package.
"""


from collections import OrderedDict
import yaml
import yamlloader
import jsonschema
import zhmcclient

from zhmcclient._utils import datetime_from_isoformat, repr_obj_id

from ._hmc import FakedHmc, FakedMetricObjectValues
from ._urihandler import UriHandler, HTTPError, URIS
from ._urihandler import ConnectionError  # pylint: disable=redefined-builtin

__all__ = ['FakedSession', 'HmcDefinitionYamlError', 'HmcDefinitionSchemaError']


# JSON schema for a faked HMC definition
FAKED_HMC_DEFINITION_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "description": "JSON schema for a faked HMC definition",
    "definitions": {
        "Properties": {
            "description": "Dictionary of resource properties. Keys are the "
                           "property names in HMC format (with dashes)",
            "type": "object",
            "patternProperties": {
                "^[a-z0-9\\-]+$": {
                    "description": "A resource property value",
                    "type": ["object", "array", "string", "integer", "number",
                             "boolean", "null"],
                },
            },
        },
        "Hmc": {
            "description": "The definition of a faked HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "host",
                "api_version",
                "consoles",
            ],
            "properties": {
                "host": {
                    "description": "The hostname or IP address of the HMC host",
                    "type": "string",
                },
                "api_version": {
                    "description": "The version of the HMC WS API, as "
                                   "major.minor",
                    "type": "string",
                },
                "metric_values": {
                    "description": "The metric values prepared for later "
                                   "retrieval",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/MetricValues"
                    },
                },
                "metrics_contexts": {
                    "description": "The metrics contexts defined on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/MetricsContext"
                    },
                },
                "consoles": {
                    "description": "The consoles (HMCs). There is only "
                                   "a single console.",
                    "type": "array",
                    "maxItems": 1,
                    "items": {
                        "$ref": "#/definitions/Console"
                    },
                },
                "cpcs": {
                    "description": "The CPCs managed by this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Cpc"
                    },
                },
            },
        },
        "MetricValues": {
            "description": "The metric values of a single metric group for a "
                           "single resource object at a point in time, "
                           "prepared for later retrieval",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "group_name",
                "resource_uri",
                "timestamp",
                "metrics",
            ],
            "properties": {
                "group_name": {
                    "description": "Name of the metric group definition for "
                                   "these metric values",
                    "type": "string",
                },
                "resource_uri": {
                    "description": "URI of the resource object for these "
                                   "metric values",
                    "type": "string",
                },
                "timestamp": {
                    "description": "Point in time for these metric values, "
                                   "as a string in ISO8601 format",
                    "type": "string",
                },
                "metrics": {
                    "description": "The metrics (values by name)",
                    "type": "object",
                    "patternProperties": {
                        "^[a-z0-9\\-]+$": {
                            "description": "The value of the metric",
                            "type": ["string", "integer", "number",
                                     "boolean", "null"],
                        },
                    },
                },
            },
        },
        "MetricsContext": {
            "description": "A metrics context defined on an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "Console": {
            "description": "A console (HMC)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
                "users": {
                    "description": "The users defined on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/User"
                    },
                },
                "user_roles": {
                    "description": "The user roles defined on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/UserRole"
                    },
                },
                "user_patterns": {
                    "description": "The user patterns defined on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/UserPattern"
                    },
                },
                "password_rules": {
                    "description": "The password rules defined on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/PasswordRule"
                    },
                },
                "tasks": {
                    "description": "The tasks defined on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Task"
                    },
                },
                "ldap_server_definitions": {
                    "description": "The LDAP server definitions on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/LdapServerDefinition"
                    },
                },
                "unmanaged_cpcs": {
                    "description": "The unmanaged CPCs discovered by this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/UnmanagedCpc"
                    },
                },
                "storage_groups": {
                    "description": "The storage groups defined on this HMC",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/StorageGroup"
                    },
                },
            },
        },
        "User": {
            "description": "A user defined on an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "UserRole": {
            "description": "A user role defined on an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "UserPattern": {
            "description": "A user pattern defined on an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "PasswordRule": {
            "description": "A password rule defined on an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "Task": {
            "description": "A task defined on an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "LdapServerDefinition": {
            "description": "An LPAP server definition on an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "UnmanagedCpc": {
            "description": "An unmanaged CPC discovered by an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "StorageGroup": {
            "description": "A storage group defined on an HMC (and associated "
                           "with a CPC)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
                "storage_volumes": {
                    "description": "The storage volumes of this storage group",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/StorageVolume"
                    },
                },
                "virtual_storage_resources": {
                    "description": "The VSRs of this storage group",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/VirtualStorageResource"
                    },
                },
            },
        },
        "StorageVolume": {
            "description": "A storage volume of a storage group",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "VirtualStorageResource": {
            "description": "A VSR of a storage group",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "Cpc": {
            "description": "A CPC managed by an HMC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
                "capacity_groups": {
                    "description": "The capacity groups of this CPC (any mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/CapacityGroup"
                    },
                },
                "partitions": {
                    "description": "The partitions of this CPC (DPM mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Partition"
                    },
                },
                "adapters": {
                    "description": "The adapters of this CPC (DPM mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Adapter"
                    },
                },
                "virtual_switches": {
                    "description": "The virtual switches of this CPC "
                                   "(DPM mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/VirtualSwitch"
                    },
                },
                "lpars": {
                    "description": "The LPARs of this CPC (classic mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Lpar"
                    },
                },
                "reset_activation_profiles": {
                    "description": "The reset activation profiles of this CPC "
                                   "(classic mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/ResetActivationProfile"
                    },
                },
                "image_activation_profiles": {
                    "description": "The image activation profiles of this CPC "
                                   "(classic mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/ImageActivationProfile"
                    },
                },
                "load_activation_profiles": {
                    "description": "The load activation profiles of this CPC "
                                   "(classic mode)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/LoadActivationProfile"
                    },
                },
            },
        },
        "CapacityGroup": {
            "description": "A capacity group in a CPC",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "Partition": {
            "description": "A partition of a CPC (DPM mode)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
                "devno_pool": {
                    "description": "Internal state: The pool of "
                                   "auto-allocated device numbers for this "
                                   "partition",
                    "type": "object",
                    "additionalProperties": True,
                },
                "wwpn_pool": {
                    "description": "Internal state: The pool of "
                                   "auto-allocated WWPNs for this partition",
                    "type": "object",
                    "additionalProperties": True,
                },
                "nics": {
                    "description": "The NICs of this partition",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Nic"
                    },
                },
                "hbas": {
                    "description": "The HBAs of this partition (up to z13)",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Hba"
                    },
                },
                "virtual_functions": {
                    "description": "The virtual functions of this partition",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/VirtualFunction"
                    },
                },
            },
        },
        "Nic": {
            "description": "A NIC of a partition",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "Hba": {
            "description": "An HBA of a partition (up to z13)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "VirtualFunction": {
            "description": "A virtual function of a partition",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "Adapter": {
            "description": "An adapter of a CPC (DPM mode)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
                "ports": {
                    "description": "The ports of this adapter",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Port"
                    },
                },
            },
        },
        "Port": {
            "description": "A port of an adapter",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "Lpar": {
            "description": "An LPAR of a CPC (classic mode)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "VirtualSwitch": {
            "description": "A virtual switch in a CPC (DPM mode)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "ResetActivationProfile": {
            "description": "A reset activation profile of a CPC (classic mode)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "ImageActivationProfile": {
            "description": "An image activation profile of a CPC "
                           "(classic mode)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
        "LoadActivationProfile": {
            "description": "A load activation profile of a CPC (classic mode)",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "properties",
            ],
            "properties": {
                "properties": {
                    "$ref": "#/definitions/Properties"
                },
            },
        },
    },
    "type": "object",
    "additionalProperties": False,
    "required": [
        "hmc_definition",
    ],
    "properties": {
        "hmc_definition": {
            "$ref": "#/definitions/Hmc"
        },
    },
}


class HmcDefinitionYamlError(Exception):
    """
    An error that is raised when loading an HMC definition and that indicates
    invalid YAML syntax in the faked HMC definition, at the YAML scanner or
    parser level.

    ``args[0]`` will be set to a message detailing the issue.
    """

    def __init__(self, message):
        # pylint: disable=useless-super-delegation
        super().__init__(message)


class HmcDefinitionSchemaError(Exception):
    """
    An error that is raised when loading an HMC definition and that indicates
    that the data in the faked HMC definition fails schema validation.

    ``args[0]`` will be set to a message detailing the issue.
    """

    def __init__(self, message):
        # pylint: disable=useless-super-delegation
        super().__init__(message)


class FakedSession(zhmcclient.Session):
    """
    A faked Session class for the zhmcclient package, that can be used as a
    replacement for the :class:`zhmcclient.Session` class.

    This class is derived from :class:`zhmcclient.Session`.

    This class can be used by projects using the zhmcclient package for their
    unit testing. It can also be used by unit tests of the zhmcclient package
    itself.

    This class provides a faked HMC with all of its resources that are relevant
    for the zhmcclient.

    The faked HMC provided by this class maintains its resource state in memory
    as Python objects, and no communication happens to any real HMC. The
    faked HMC implements all HMC operations that are relevant for the
    zhmcclient package in a successful manner.

    It is possible to populate the faked HMC with an initial resource state
    (see :meth:`~zhmcclient_mock.FakedHmc.add_resources`).
    """

    def __init__(self, host, hmc_name, hmc_version, api_version,
                 userid=None, password=None):
        """
        Parameters:

          host (:term:`string`):
            HMC host the mocked HMC will be set up with.

          hmc_name (:term:`string`):
            HMC name. Used for result of Query Version Info operation.

          hmc_version (:term:`string`):
            HMC version string (e.g. '2.13.1'). Used for result of
            Query Version Info operation.

          api_version (:term:`string`):
            HMC API version string (e.g. '1.8'). Used for result of
            Query Version Info operation.

          userid (:term:`string`):
            HMC userid for logging in to the mocked HMC.

          password (:term:`string`):
            HMC password for logging in to the mocked HMC.
        """
        super().__init__(
            host, userid=userid, password=password)
        self._hmc = FakedHmc(self, hmc_name, hmc_version, api_version)
        self._urihandler = UriHandler(URIS)
        self._object_topic = 'faked-notification-topic'
        self._job_topic = 'faked-job-notification-topic'

    def __repr__(self):
        """
        Return a string with the state of this faked session, for debug
        purposes.
        """
        ret = (
            f"{repr_obj_id(self)} (\n"
            f"  _hosts = {self._hosts!r}\n"
            f"  _userid = {self._userid!r}\n"
            f"  _password = '...'\n"
            f"  _get_password = {self._get_password!r}\n"
            f"  _retry_timeout_config = {self._retry_timeout_config!r}\n"
            f"  _actual_host = {self._actual_host!r}\n"
            f"  _base_url = {self._base_url!r}\n"
            f"  _headers = {self._headers!r}\n"
            f"  _session_id = {self._session_id!r}\n"
            f"  _session = {self._session!r}\n"
            f"  _hmc = {repr_obj_id(self._hmc)}\n"
            f"  _urihandler = {self._urihandler!r}\n"
            ")")
        return ret

    @property
    def hmc(self):
        """
        :class:`~zhmcclient_mock.FakedHmc`: The faked HMC provided by this
        faked session.

        The faked HMC supports being populated with initial resource state,
        for example using its :meth:`zhmcclient_mock.FakedHmc.add_resources`
        method.

        As an alternative to providing an entire resource tree, the resources
        can also be added one by one, from top to bottom, using the
        :meth:`zhmcclient_mock.FakedBaseManager.add` methods of the
        respective managers (the top-level manager for CPCs can be accessed
        via ``hmc.cpcs``).
        """
        return self._hmc

    @staticmethod
    def from_hmc_yaml_file(filepath, userid=None, password=None):
        """
        Return a new FakedSession object from an HMC definition in a YAML file.

        The data format of the YAML file is validated using a schema.

        Parameters:

          filepath(:term:`string`): Path name of the YAML file that contains
            the HMC definition.

          userid (:term:`string`):
            Userid of the HMC user to be used for logging in, or `None`.

          password (:term:`string`):
            Password of the HMC user if `userid` was specified, or `None`.

        Returns:

          FakedSession: New faked session with faked HMC set up from HMC
          definition.

        Raises:

            IOError: Error opening the YAML file for reading.
            YamlFormatError: Invalid YAML syntax in HMC definition.
            HmcDefinitionSchemaError: Invalid data format in HMC definition.
        """
        # pylint: disable=unspecified-encoding
        with open(filepath) as fp:
            hmc = FakedSession.from_hmc_yaml(fp, filepath, userid, password)
        return hmc

    @staticmethod
    def from_hmc_yaml(hmc_yaml, filepath=None, userid=None, password=None):
        """
        Return a new FakedSession object from an HMC definition YAML string
        or stream.

        An HMC definition YAML string can be created using
        :meth:`zhmcclient.Client.to_hmc_yaml`.

        The timestamp in metric values can have any valid ISO8601 format.
        Timezone-naive values are amended with the local timezone.

        The data format of the YAML string is validated using a schema.

        Parameters:

          hmc_yaml(string or stream): HMC definition YAML string or stream.

          filepath(string): Path name of the YAML file that contains the HMC
            definition; used only in exception messages. If `None`, no
            filename is used in exception messages.

          userid (:term:`string`):
            Userid of the HMC user to be used for logging in, or `None`.

          password (:term:`string`):
            Password of the HMC user if `userid` was specified, or `None`.

        Returns:

          FakedSession: New faked session with faked HMC set up from HMC
          definition.

        Raises:

            YamlFormatError: Invalid YAML syntax in HMC definition YAML string
              or stream.
            HmcDefinitionSchemaError: Invalid data format in HMC definition.
        """

        try:
            hmc_dict = yaml.load(
                hmc_yaml, Loader=yamlloader.ordereddict.SafeLoader)
        except (yaml.parser.ParserError, yaml.scanner.ScannerError) as exc:
            if filepath:
                file_str = f" in file {filepath}"
            else:
                file_str = ""
            new_exc = HmcDefinitionYamlError(
                f"Invalid YAML syntax in faked HMC definition{file_str}: {exc}")
            new_exc.__cause__ = None
            raise new_exc  # HmcDefinitionYamlError

        hmc = FakedSession.from_hmc_dict(hmc_dict, filepath, userid, password)
        return hmc

    @staticmethod
    def from_hmc_dict(hmc_dict, filepath=None, userid=None, password=None):
        """
        Return a new FakedSession object from an HMC definition dictionary.

        An HMC definition dictionary can be created using
        :meth:`zhmcclient.Client.to_hmc_dict`.

        The timestamp in metric values can have any valid ISO8601 format.
        Timezone-naive values are amended with the local timezone.

        The data format of the YAML string is validated using a schema.

        Parameters:

          hmc_dict(dict): HMC definition dictionary.

          filepath(string): Path name of the YAML file that contains the HMC
            definition; used only in exception messages. If `None`, no
            filename is used in exception messages.

          userid (:term:`string`):
            Userid of the HMC user to be used for logging in, or `None`.

          password (:term:`string`):
            Password of the HMC user if `userid` was specified, or `None`.

        Returns:

          FakedSession: New faked session with faked HMC set up from the HMC
          definition.

        Raises:

            HmcDefinitionSchemaError: Invalid data format in HMC definition.
        """

        try:
            jsonschema.validate(hmc_dict, FAKED_HMC_DEFINITION_SCHEMA)
        except jsonschema.exceptions.ValidationError as exc:
            if filepath:
                file_str = f" in file {filepath}"
            else:
                file_str = ""
            elem = '.'.join(str(e) for e in exc.absolute_path)
            schemaitem = '.'.join(str(e) for e in exc.absolute_schema_path)
            new_exc = HmcDefinitionSchemaError(
                f"Invalid data format in faked HMC definition{file_str}: "
                f"{exc.message}; "
                f"Offending element: {elem}; "
                f"Schema item: {schemaitem}; "
                f"Validator: {exc.validator}={exc.validator_value}")
            new_exc.__cause__ = None
            raise new_exc  # HmcDefinitionSchemaError

        hmc_res_dict = hmc_dict['hmc_definition']

        consoles = hmc_res_dict.get('consoles')
        console = consoles[0]
        host = hmc_res_dict['host']
        api_version = hmc_res_dict['api_version']
        hmc_name = console['properties']['name']
        hmc_version = console['properties']['version']

        session = FakedSession(host, hmc_name, hmc_version, api_version,
                               userid=userid, password=password)

        res_dict = OrderedDict()
        res_dict['consoles'] = consoles
        cpcs = hmc_res_dict.get('cpcs')
        if cpcs:
            res_dict['cpcs'] = cpcs
        metrics_contexts = hmc_res_dict.get('metrics_contexts')
        if metrics_contexts:
            res_dict['metrics_contexts'] = metrics_contexts
        session.hmc.add_resources(res_dict)

        mv_dicts = hmc_res_dict.get('metric_values')
        if mv_dicts:
            for mv_dict in mv_dicts:
                group_name = mv_dict['group_name']
                resource_uri = mv_dict['resource_uri']
                timestamp = datetime_from_isoformat(mv_dict['timestamp'])
                values = []
                for name, value in mv_dict['metrics'].items():
                    item_tup = (name, value)
                    values.append(item_tup)
                mv = FakedMetricObjectValues(
                    group_name=group_name,
                    resource_uri=resource_uri,
                    timestamp=timestamp,
                    values=values)
                session.hmc.add_metric_values(mv)

        return session

    def get(self, uri, resource=None, logon_required=True, renew_session=True):
        """
        Perform the HTTP GET method against the resource identified by a URI,
        on the faked HMC.

        Parameters:

          uri (:term:`string`):
            Relative URI path of the resource, e.g. "/api/session".
            This URI is relative to the base URL of the session (see
            the :attr:`~zhmcclient.Session.base_url` property).
            Must not be `None`.

          logon_required (bool):
            Boolean indicating whether the operation requires that the session
            is logged on to the HMC.

            Because this is a faked HMC, this does not perform a real logon,
            but it is still used to update the state in the faked HMC.

          renew_session (bool):
            Boolean indicating whether the session should be renewed in case
            it is expired.

            This parameter exists for compatibility with real HMCs, but is
            ignored.

        Returns:

          :term:`json object` with the operation result.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError` (not implemented)
          :exc:`~zhmcclient.AuthError` (not implemented)
          :exc:`~zhmcclient.ConnectionError`
        """
        try:
            return self._urihandler.get(self._hmc, uri, logon_required)
        except HTTPError as exc:
            new_exc = zhmcclient.HTTPError(exc.response())
            new_exc.__cause__ = None
            raise new_exc  # zhmcclient.HTTPError
        except ConnectionError as exc:
            new_exc = zhmcclient.ConnectionError(exc.message, None)
            new_exc.__cause__ = None
            raise new_exc  # zhmcclient.ConnectionError

    def post(self, uri, resource=None, body=None, logon_required=True,
             wait_for_completion=True, operation_timeout=None,
             renew_session=True):
        """
        Perform the HTTP POST method against the resource identified by a URI,
        using a provided request body, on the faked HMC.

        HMC operations using HTTP POST are either synchronous or asynchronous.
        Asynchronous operations return the URI of an asynchronously executing
        job that can be queried for status and result.

        Examples for synchronous operations:

        * With no response body: "Logon", "Update CPC Properties"
        * With a response body: "Create Partition"

        Examples for asynchronous operations:

        * With no ``job-results`` field in the completed job status response:
          "Start Partition"
        * With a ``job-results`` field in the completed job status response
          (under certain conditions): "Activate a Blade", or "Set CPC Power
          Save"

        The `wait_for_completion` parameter of this method can be used to deal
        with asynchronous HMC operations in a synchronous way.

        Parameters:

          uri (:term:`string`):
            Relative URI path of the resource, e.g. "/api/session".
            This URI is relative to the base URL of the session (see the
            :attr:`~zhmcclient.Session.base_url` property).
            Must not be `None`.

          body (:term:`json object`):
            JSON object to be used as the HTTP request body (payload).
            `None` means the same as an empty dictionary, namely that no HTTP
            body is included in the request.

          logon_required (bool):
            Boolean indicating whether the operation requires that the session
            is logged on to the HMC. For example, the "Logon" operation does
            not require that.

            Because this is a faked HMC, this does not perform a real logon,
            but it is still used to update the state in the faked HMC.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested HMC operation, as follows:

            * If `True`, this method will wait for completion of the requested
              operation, regardless of whether the operation is synchronous or
              asynchronous.

              This will cause an additional entry in the time statistics to be
              created for the asynchronous operation and waiting for its
              completion. This entry will have a URI that is the targeted URI,
              appended with "+completion".

            * If `False`, this method will immediately return the result of the
              HTTP POST method, regardless of whether the operation is
              synchronous or asynchronous.

          operation_timeout (:term:`number`):
            Timeout in seconds, when waiting for completion of an asynchronous
            operation. The special value 0 means that no timeout is set. `None`
            means that the default async operation timeout of the session is
            used.

            For `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised when the timeout
            expires.

            For `wait_for_completion=False`, this parameter has no effect.

          renew_session (bool):
            Boolean indicating whether the session should be renewed in case
            it is expired.

            This parameter exists for compatibility with real HMCs, but is
            ignored.

        Returns:

          :term:`json object`:

            If `wait_for_completion` is `True`, returns a JSON object
            representing the response body of the synchronous operation, or the
            response body of the completed job that performed the asynchronous
            operation. If a synchronous operation has no response body, `None`
            is returned.

            If `wait_for_completion` is `False`, returns a JSON object
            representing the response body of the synchronous or asynchronous
            operation. In case of an asynchronous operation, the JSON object
            will have a member named ``job-uri``, whose value can be used with
            the :meth:`~zhmcclient.Session.query_job_status` method to
            determine the status of the job and the result of the original
            operation, once the job has completed.

            See the section in the :term:`HMC API` book about the specific HMC
            operation and about the 'Query Job Status' operation, for a
            description of the members of the returned JSON objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError` (not implemented)
          :exc:`~zhmcclient.AuthError` (not implemented)
          :exc:`~zhmcclient.ConnectionError`
        """
        try:
            return self._urihandler.post(self._hmc, uri, body, logon_required,
                                         wait_for_completion)
        except HTTPError as exc:
            new_exc = zhmcclient.HTTPError(exc.response())
            new_exc.__cause__ = None
            raise new_exc  # zhmcclient.HTTPError
        except ConnectionError as exc:
            new_exc = zhmcclient.ConnectionError(exc.message, None)
            new_exc.__cause__ = None
            raise new_exc  # zhmcclient.ConnectionError

    def delete(
            self, uri, resource=None, logon_required=True, renew_session=True):
        """
        Perform the HTTP DELETE method against the resource identified by a
        URI, on the faked HMC.

        Parameters:

          uri (:term:`string`):
            Relative URI path of the resource, e.g.
            "/api/session/{session-id}".
            This URI is relative to the base URL of the session (see
            the :attr:`~zhmcclient.Session.base_url` property).
            Must not be `None`.

          logon_required (bool):
            Boolean indicating whether the operation requires that the session
            is logged on to the HMC. For example, for the logoff operation, it
            does not make sense to first log on.

            Because this is a faked HMC, this does not perform a real logon,
            but it is still used to update the state in the faked HMC.

          renew_session (bool):
            Boolean indicating whether the session should be renewed in case
            it is expired.

            This parameter exists for compatibility with real HMCs, but is
            ignored.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError` (not implemented)
          :exc:`~zhmcclient.AuthError` (not implemented)
          :exc:`~zhmcclient.ConnectionError`
        """
        try:
            self._urihandler.delete(self._hmc, uri, logon_required)
        except HTTPError as exc:
            new_exc = zhmcclient.HTTPError(exc.response())
            new_exc.__cause__ = None
            raise new_exc  # zhmcclient.HTTPError
        except ConnectionError as exc:
            new_exc = zhmcclient.ConnectionError(exc.message, None)
            new_exc.__cause__ = None
            raise new_exc  # zhmcclient.ConnectionError



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Capacity Group` is a group of :term:`partitions <Partition>` that
allows defining a capping for the total amount of physical processors consumed
by the Partitions in the group. The Partitions must be defined with shared
processors, and a Partition can be a member of at most one Capacity Group.

Capacity Group resources are contained in CPC resources.

Capacity Groups only exist in :term:`CPCs <CPC>` that are in DPM mode.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_CAPACITY_GROUP

__all__ = ['CapacityGroupManager', 'CapacityGroup']


class CapacityGroupManager(BaseManager):
    """
    Manager providing access to the :term:`Capacity Groups <Capacity Group>`
    in a particular :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Cpc` object (in DPM mode):

    * :attr:`~zhmcclient.Cpc.capacity_groups`

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, cpc):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.

        super().__init__(
            resource_class=CapacityGroup,
            class_name=RC_CAPACITY_GROUP,
            session=cpc.manager.session,
            parent=cpc,
            base_uri=f'{cpc.uri}/capacity-groups',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=[],
            list_has_name=False)

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope
        for this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the Capacity Groups in this CPC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.CapacityGroup` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'capacity-groups'
        list_uri = f'{self.cpc.uri}/capacity-groups'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties):
        """
        Create a Capacity Group in this CPC.

        The new Capacity Group initially has no partitions.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.
        * Task permission to the "Manage Processor Sharing" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Capacity Group' in the :term:`HMC API` book.

        Returns:

          CapacityGroup:
            The resource object for the new Capacity Group.
            The object will have its 'element-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.cpc.uri + '/capacity-groups',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        capacity_group = CapacityGroup(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return capacity_group


class CapacityGroup(BaseResource):
    """
    Representation of a :term:`Capacity Group`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of a Capacity Group resource, see section
    'Data model' in section 'Capacity Group element object' in the
    :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.CapacityGroupManager`).
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.CapacityGroupManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, CapacityGroupManager), (
            "CapacityGroup init: Expected manager type "
            f"{CapacityGroupManager}, got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this Capacity Group.

        The Capacity Group must not contain any Partitions.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the CPC containing this Capacity Group.
        * Task permission to the "Manage Processor Sharing" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self._uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Capacity Group.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the CPC containing this Capacity Group.
        * Task permission to the "Manage Processor Sharing" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Capacity Group element object'
            in the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    @logged_api_call
    def add_partition(self, partition):
        """
        Add a Partition to this Capacity Group.

        Upon successful completion, the amount of processing capacity that
        could be used by this Partition becomes governed by the absolute
        cap values defined for this Capacity Group.

        A Partition cannot become a member of more than one Capacity Group.
        The Partition must be defined with shared processors and the
        Capacity Group must specify nonzero cap values for the processor types
        used by the Partition. The Partition must be on the same CPC as the
        Capacity Group and must not yet be a member of this (or any other)
        Capacity Group.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the CPC containing this Capacity Group.
        * Task permission to the "Manage Processor Sharing" task.

        Parameters:

          partition (:class:`~zhmcclient.Partition`): The partition to be added.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'partition-uri': partition.uri,
        }
        self.manager.session.post(
            self.uri + '/operations/add-partition', resource=self,
            body=body)

    @logged_api_call
    def remove_partition(self, partition):
        """
        Remove a Partition from this Capacity Group.

        Upon successful completion, the amount of processing capacity that
        could be used by this Partition is no longer governed by the absolute
        cap values defined for this Capacity Group.

        The Partition must be a member of this Capacity Group.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the CPC containing this Capacity Group.
        * Task permission to the "Manage Processor Sharing" task.

        Parameters:

          partition (:class:`~zhmcclient.Partition`): The partition to be
            removed.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'partition-uri': partition.uri,
        }
        self.manager.session.post(
            self.uri + '/operations/remove-partition', resource=self,
            body=body)



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _user module.
"""


import re
import copy
import logging
import pytest

from zhmcclient import Client, HTTPError, NotFound, User
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources, assert_blanked_in_message


class TestUser:
    """All tests for the User and UserManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked Console without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_console = self.session.hmc.consoles.add({
            'object-id': None,
            # object-uri will be automatically set
            'parent': None,
            'class': 'console',
            'name': 'fake-console1',
            'description': 'Console #1',
        })
        self.console = self.client.consoles.find(name=self.faked_console.name)

    def add_user(self, name, type_):
        """
        Add a faked user object to the faked Console and return it.
        """
        faked_user = self.faked_console.users.add({
            'object-id': f'oid-{name}',
            # object-uri will be automatically set
            'parent': '/api/console',
            'class': 'user',
            'name': name,
            'description': f'User {name}',
            'type': type_,
            'authentication-type': 'local',
        })
        return faked_user

    def add_user_role(self, name, type_):
        """
        Add a faked user role object to the faked Console and return it.
        """
        faked_user_role = self.faked_console.user_roles.add({
            'object-id': f'oid-{name}',
            # object-uri will be automatically set
            'parent': '/api/console',
            'class': 'user-role',
            'name': name,
            'description': f'User Role {name}',
            'type': type_,
        })
        return faked_user_role

    def test_user_manager_repr(self):
        """Test UserManager.__repr__()."""

        user_mgr = self.console.users

        # Execute the code to be tested
        repr_str = repr(user_mgr)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{user_mgr.__class__.__name__}\s+at\s+'
            rf'0x{id(user_mgr):08x}\s+\(\\n.*',
            repr_str)

    def test_user_manager_initial_attrs(self):
        """Test initial attributes of UserManager."""

        user_mgr = self.console.users

        # Verify all public properties of the manager object
        assert user_mgr.resource_class == User
        assert user_mgr.class_name == 'user'
        assert user_mgr.session is self.session
        assert user_mgr.parent is self.console
        assert user_mgr.console is self.console

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            (dict(full_properties=False),
             ['object-uri', 'name', 'type']),
            (dict(full_properties=True),
             ['object-uri', 'name', 'type', 'description']),
            ({},  # test default for full_properties (False)
             ['object-uri', 'name', 'type']),
        ]
    )
    @pytest.mark.parametrize(
        "filter_args, exp_names", [
            (None,
             ['a', 'b']),
            ({},
             ['a', 'b']),
            ({'name': 'a'},
             ['a']),
            ({'name': 'A'},  # users have case-insensitive names
             ['a']),
        ]
    )
    def test_user_manager_list(
            self, filter_args, exp_names, full_properties_kwargs, prop_names):
        """Test UserManager.list()."""

        faked_user1 = self.add_user(name='a', type_='standard')
        faked_user2 = self.add_user(name='b', type_='standard')
        faked_users = [faked_user1, faked_user2]
        exp_faked_users = [u for u in faked_users if u.name in exp_names]
        user_mgr = self.console.users

        # Execute the code to be tested
        users = user_mgr.list(filter_args=filter_args,
                              **full_properties_kwargs)

        assert_resources(users, exp_faked_users, prop_names)

    @pytest.mark.parametrize(
        "input_props, exp_prop_names, exp_exc", [
            ({},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'description': 'fake description X'},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-name-x'},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-name-x',
              'type': 'standard'},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-name-x',
              'type': 'standard',
              'authentication-type': 'local',
              'password': 'bla',
              'password-rule-uri': '/api/console/password-rules/dummy'},
             ['object-uri', 'name', 'type', 'authentication-type'],
             None),
            ({'name': 'fake-name-x',
              'type': 'standard',
              'authentication-type': 'local',
              'password': 'bla',
              'password-rule-uri': '/api/console/password-rules/dummy',
              'description': 'fake description X'},
             ['object-uri', 'name', 'type', 'authentication-type',
              'description'],
             None),
        ]
    )
    def test_user_manager_create(
            self, caplog, input_props, exp_prop_names, exp_exc):
        """Test UserManager.create()."""

        logger_name = "zhmcclient.api"
        caplog.set_level(logging.DEBUG, logger=logger_name)

        user_mgr = self.console.users

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user_mgr.create(properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            user = user_mgr.create(properties=input_props)

            # Get its API call log record
            call_record = caplog.records[-2]

            # Check the resource for consistency within itself
            assert isinstance(user, User)
            user_name = user.name
            exp_user_name = user.properties['name']
            assert user_name == exp_user_name
            user_uri = user.uri
            exp_user_uri = user.properties['object-uri']
            assert user_uri == exp_user_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in user.properties
                if prop_name in input_props:
                    value = user.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

            # Verify the API call log record for blanked-out properties.
            assert_blanked_in_message(
                call_record.message, input_props,
                ['password'])

    def test_user_repr(self):
        """Test User.__repr__()."""

        faked_user1 = self.add_user(name='a', type_='standard')
        user1 = self.console.users.find(name=faked_user1.name)

        # Execute the code to be tested
        repr_str = repr(user1)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{user1.__class__.__name__}\s+at\s+'
            rf'0x{id(user1):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "input_name, input_type, exp_exc", [
            ('a', 'standard', None),
            ('b', 'template', None),
            ('c', 'pattern-based',
             HTTPError({'http-status': 400, 'reason': 312})),
            ('d', 'system-defined', None),
        ]
    )
    def test_user_delete(self, input_name, input_type, exp_exc):
        """Test User.delete()."""

        faked_user = self.add_user(name=input_name, type_=input_type)
        user_mgr = self.console.users
        user = user_mgr.find(name=faked_user.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user.delete()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the user still exists
            user_mgr.find(name=faked_user.name)

        else:

            # Execute the code to be tested.
            user.delete()

            # Check that the user no longer exists
            with pytest.raises(NotFound) as exc_info:
                user_mgr.find(name=faked_user.name)

    def test_user_delete_create_same_name(self):
        """Test User.delete() followed by create() with same name."""

        user_name = 'faked_a'

        # Add the user to be tested
        self.add_user(name=user_name, type_='standard')

        # Input properties for a user with the same name
        sn_user_props = {
            'name': user_name,
            'description': 'User with same name',
            'type': 'standard',
            'authentication-type': 'local',
            'password': 'bla',
            'password-rule-uri': '/api/console/password-rules/dummy',
        }

        user_mgr = self.console.users
        user = user_mgr.find(name=user_name)

        # Execute the deletion code to be tested
        user.delete()

        # Check that the user no longer exists
        with pytest.raises(NotFound):
            user_mgr.find(name=user_name)

        # Execute the creation code to be tested.
        user_mgr.create(sn_user_props)

        # Check that the user exists again under that name
        sn_user = user_mgr.find(name=user_name)
        description = sn_user.get_property('description')
        assert description == sn_user_props['description']

    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New user description'},
            {'authentication-type': 'ldap',
             'description': 'New user description'},
        ]
    )
    def test_user_update_properties(self, input_props):
        """Test User.update_properties()."""

        # Add the user to be tested
        faked_user = self.add_user(name='a', type_='standard')

        user_mgr = self.console.users
        user = user_mgr.find(name=faked_user.name)

        user.pull_full_properties()
        saved_props = copy.deepcopy(user.properties)

        # Execute the code to be tested
        user.update_properties(properties=input_props)

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_props:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_props[prop_name]
            assert prop_name in user.properties
            prop_value = user.properties[prop_name]
            assert prop_value == exp_prop_value, \
                f"Unexpected value for property {prop_name!r}"

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        user.pull_full_properties()
        for prop_name in saved_props:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_props[prop_name]
            assert prop_name in user.properties
            prop_value = user.properties[prop_name]
            assert prop_value == exp_prop_value

    @pytest.mark.parametrize(
        "user_name, user_type, exp_exc", [
            ('a', 'standard', None),
            ('b', 'template', None),
            ('c', 'pattern-based',
             HTTPError({'http-status': 400, 'reason': 314})),
            ('d', 'system-defined',
             HTTPError({'http-status': 400, 'reason': 314})),
        ]
    )
    @pytest.mark.parametrize(
        "role_name, role_type", [
            ('ra', 'user-defined'),
            ('rb', 'system-defined'),
        ]
    )
    def test_user_add_user_role(
            self, role_name, role_type, user_name, user_type, exp_exc):
        """Test User.add_user_role()."""

        faked_user = self.add_user(name=user_name, type_=user_type)
        user_mgr = self.console.users
        user = user_mgr.find(name=faked_user.name)

        faked_user_role = self.add_user_role(name=role_name, type_=role_type)
        user_role_mgr = self.console.user_roles
        user_role = user_role_mgr.find(name=faked_user_role.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user.add_user_role(user_role)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the user does not have that user role
            user.pull_full_properties()
            if 'user-roles' in user.properties:
                user_role_uris = user.properties['user-roles']
                user_role_uri = user_role.uri
                assert user_role_uri not in user_role_uris

        else:

            # Execute the code to be tested.
            ret = user.add_user_role(user_role)

            assert ret is None

            # Check that the user has that user role
            user.pull_full_properties()
            assert 'user-roles' in user.properties
            user_role_uris = user.properties['user-roles']
            user_role_uri = user_role.uri
            assert user_role_uri in user_role_uris

    @pytest.mark.parametrize(
        "user_name, user_type, exp_exc", [
            ('a', 'standard', None),
            ('b', 'template', None),
            ('c', 'pattern-based',
             HTTPError({'http-status': 400, 'reason': 314})),
            ('d', 'system-defined',
             HTTPError({'http-status': 400, 'reason': 314})),
        ]
    )
    @pytest.mark.parametrize(
        "role_name, role_type", [
            ('ra', 'user-defined'),
            ('rb', 'system-defined'),
        ]
    )
    def test_user_remove_user_role(
            self, role_name, role_type, user_name, user_type, exp_exc):
        """Test User.remove_user_role()."""

        faked_user = self.add_user(name=user_name, type_=user_type)
        user_mgr = self.console.users
        user = user_mgr.find(name=faked_user.name)

        faked_user_role = self.add_user_role(name=role_name, type_=role_type)
        user_role_mgr = self.console.user_roles
        user_role = user_role_mgr.find(name=faked_user_role.name)

        # Prepare the user with the initial user role
        if 'user-roles' not in faked_user.properties:
            faked_user.properties['user-roles'] = []
        faked_user.properties['user-roles'].append(faked_user_role.uri)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user.remove_user_role(user_role)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the user still has that user role
            user.pull_full_properties()
            if 'user-roles' in user.properties:
                user_role_uris = user.properties['user-roles']
                user_role_uri = user_role.uri
                assert user_role_uri in user_role_uris

        else:

            # Execute the code to be tested.
            ret = user.remove_user_role(user_role)

            assert ret is None

            # Check that the user no longer has that user role
            user.pull_full_properties()
            assert 'user-roles' in user.properties
            user_role_uris = user.properties['user-roles']
            user_role_uri = user_role.uri
            assert user_role_uri not in user_role_uris



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _hba module.
"""


import re
import copy
import pytest

from zhmcclient import Client, Hba, HTTPError, NotFound
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


# Object IDs and names of our faked HBAs:
HBA1_OID = 'hba 1-oid'
HBA1_NAME = 'hba 1'
HBA2_OID = 'hba 2-oid'
HBA2_NAME = 'hba 2'

# URIs and Object IDs of elements referenced in HBA properties:
FCP1_OID = 'fake-fcp1-oid'
PORT11_OID = 'fake-port11-oid'
PORT11_URI = f'/api/adapters/{FCP1_OID}/storage-ports/{PORT11_OID}'


class TestHba:
    """All tests for Hba and HbaManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked CPC in DPM mode with one
        partition that has no HBAs.
        Add one FCP adapter and port.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        # Add a CPC in DPM mode
        self.faked_cpc = self.session.hmc.cpcs.add({
            'element-id': 'fake-cpc1-oid',
            # element-uri is set up automatically
            'parent': None,
            'class': 'cpc',
            'name': 'fake-cpc1-name',
            'description': 'CPC #1 (DPM mode)',
            'status': 'active',
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'iml-mode': 'dpm',
        })
        self.cpc = self.client.cpcs.find(name='fake-cpc1-name')

        # Add a partition to the CPC
        self.faked_partition = self.faked_cpc.partitions.add({
            'element-id': 'fake-part1-oid',
            # element-uri will be automatically set
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': 'fake-part1-name',
            'description': 'Partition #1',
            'status': 'active',
            'initial-memory': 1024,
            'maximum-memory': 2048,
        })
        self.partition = self.cpc.partitions.find(name='fake-part1-name')

        # Add an FCP adapter and port to the CPC
        self.faked_fcp1 = self.faked_cpc.adapters.add({
            'object-id': FCP1_OID,
            'parent': self.faked_cpc.uri,
            'class': 'adapter',
            'name': 'fcp1',
            'description': 'FCP #1',
            'status': 'active',
            'type': 'fcp',
            'adapter-id': '123',
            'detected-card-type': '10gbe-roce-express',
            'card-location': '1234-5678-J.01',
            'port-count': 1,
            'network-port-uris': [],
            'state': 'online',
            'configured-capacity': 80,
            'used-capacity': 0,
            'allowed-capacity': 80,
            'maximum-total-capacity': 80,
            'physical-channel-status': 'operating',
        })
        self.faked_port11 = self.faked_fcp1.ports.add({
            'element-id': PORT11_OID,
            'parent': self.faked_fcp1.uri,
            'class': 'storage-port',
            'index': 1,
            'name': 'fake-port11-name',
            'description': 'FCP #1 Port #1',
        })
        assert PORT11_URI == self.faked_port11.uri

    def add_hba1(self):
        """Add a faked HBA 1 to the faked partition."""
        faked_hba = self.faked_partition.hbas.add({
            'element-id': HBA1_OID,
            # element-uri will be automatically set
            'parent': self.faked_partition.uri,
            'class': 'hba',
            'name': HBA1_NAME,
            'description': 'HBA ' + HBA1_NAME,
            'adapter-port-uri': PORT11_URI,
            'wwpn': 'AABBCCDDEEFF0011',
            'device-number': '1111',
        })
        return faked_hba

    def add_hba2(self):
        """Add a faked HBA 2 to the faked partition."""
        faked_hba = self.faked_partition.hbas.add({
            'element-id': HBA2_OID,
            # element-uri will be automatically set
            'parent': self.faked_partition.uri,
            'class': 'hba',
            'name': HBA2_NAME,
            'description': 'HBA ' + HBA2_NAME,
            'adapter-port-uri': PORT11_URI,
            'wwpn': 'AABBCCDDEEFF0012',
            'device-number': '1112',
        })
        return faked_hba

    def test_hbamanager_initial_attrs(self):
        """Test initial attributes of HbaManager."""

        hba_mgr = self.partition.hbas

        # Verify all public properties of the manager object
        assert hba_mgr.resource_class == Hba
        assert hba_mgr.session == self.session
        assert hba_mgr.parent == self.partition
        assert hba_mgr.partition == self.partition

    # TODO: Test for HbaManager.__repr__()

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            ({},
             ['element-uri']),
            (dict(full_properties=False),
             ['element-uri']),
            (dict(full_properties=True),
             None),
        ]
    )
    def test_hbamanager_list_full_properties(
            self, full_properties_kwargs, prop_names):
        """Test HbaManager.list() with full_properties."""

        # Add two faked HBAs
        faked_hba1 = self.add_hba1()
        faked_hba2 = self.add_hba2()

        exp_faked_hbas = [faked_hba1, faked_hba2]
        hba_mgr = self.partition.hbas

        # Execute the code to be tested
        hbas = hba_mgr.list(**full_properties_kwargs)

        assert_resources(hbas, exp_faked_hbas, prop_names)

    @pytest.mark.parametrize(
        "filter_args, exp_oids", [
            ({'element-id': HBA1_OID},
             [HBA1_OID]),
            ({'element-id': HBA2_OID},
             [HBA2_OID]),
            ({'element-id': [HBA1_OID, HBA2_OID]},
             [HBA1_OID, HBA2_OID]),
            ({'element-id': [HBA1_OID, HBA1_OID]},
             [HBA1_OID]),
            ({'element-id': HBA1_OID + 'foo'},
             []),
            ({'element-id': [HBA1_OID, HBA2_OID + 'foo']},
             [HBA1_OID]),
            ({'element-id': [HBA2_OID + 'foo', HBA1_OID]},
             [HBA1_OID]),
            ({'name': HBA1_NAME},
             [HBA1_OID]),
            ({'name': HBA2_NAME},
             [HBA2_OID]),
            ({'name': [HBA1_NAME, HBA2_NAME]},
             [HBA1_OID, HBA2_OID]),
            ({'name': HBA1_NAME + 'foo'},
             []),
            ({'name': [HBA1_NAME, HBA2_NAME + 'foo']},
             [HBA1_OID]),
            ({'name': [HBA2_NAME + 'foo', HBA1_NAME]},
             [HBA1_OID]),
            ({'name': [HBA1_NAME, HBA1_NAME]},
             [HBA1_OID]),
            ({'name': '.*hba 1'},
             [HBA1_OID]),
            ({'name': 'hba 1.*'},
             [HBA1_OID]),
            ({'name': 'hba .'},
             [HBA1_OID, HBA2_OID]),
            ({'name': '.ba 1'},
             [HBA1_OID]),
            ({'name': '.+'},
             [HBA1_OID, HBA2_OID]),
            ({'name': 'hba 1.+'},
             []),
            ({'name': '.+hba 1'},
             []),
            ({'name': HBA1_NAME,
              'element-id': HBA1_OID},
             [HBA1_OID]),
            ({'name': HBA1_NAME,
              'element-id': HBA1_OID + 'foo'},
             []),
            ({'name': HBA1_NAME + 'foo',
              'element-id': HBA1_OID},
             []),
            ({'name': HBA1_NAME + 'foo',
              'element-id': HBA1_OID + 'foo'},
             []),
        ]
    )
    def test_hbamanager_list_filter_args(self, filter_args, exp_oids):
        """Test HbaManager.list() with filter_args."""

        # Add two faked HBAs
        self.add_hba1()
        self.add_hba2()

        hba_mgr = self.partition.hbas

        # Execute the code to be tested
        hbas = hba_mgr.list(filter_args=filter_args)

        assert len(hbas) == len(exp_oids)
        if exp_oids:
            oids = [hba.properties['element-id'] for hba in hbas]
            assert set(oids) == set(exp_oids)

    @pytest.mark.parametrize(
        "initial_partition_status, exp_status_exc", [
            ('stopped', None),
            ('terminated', None),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', None),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', None),
            ('reservation-error', None),
            ('paused', None),
        ]
    )
    @pytest.mark.parametrize(
        "input_props, exp_prop_names, exp_prop_exc", [
            ({},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-hba-x'},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'adapter-port-uri': PORT11_URI},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-hba-x',
              'adapter-port-uri': PORT11_URI},
             ['element-uri', 'name', 'adapter-port-uri'],
             None),
        ]
    )
    def test_hbamanager_create(
            self, input_props, exp_prop_names, exp_prop_exc,
            initial_partition_status, exp_status_exc):
        """Test HbaManager.create()."""

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = initial_partition_status

        hba_mgr = self.partition.hbas

        if exp_status_exc:
            exp_exc = exp_status_exc
        elif exp_prop_exc:
            exp_exc = exp_prop_exc
        else:
            exp_exc = None

        if exp_exc:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                hba = hba_mgr.create(properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            # Note: the Hba object returned by Hba.create() has
            # the input properties plus 'element-uri' plus 'element-id'.
            hba = hba_mgr.create(properties=input_props)

            # Check the resource for consistency within itself
            assert isinstance(hba, Hba)
            hba_name = hba.name
            exp_hba_name = hba.properties['name']
            assert hba_name == exp_hba_name
            hba_uri = hba.uri
            exp_hba_uri = hba.properties['element-uri']
            assert hba_uri == exp_hba_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in hba.properties
                if prop_name in input_props:
                    value = hba.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

    def test_hba_repr(self):
        """Test Hba.__repr__()."""

        # Add a faked hba
        faked_hba = self.add_hba1()

        hba_mgr = self.partition.hbas
        hba = hba_mgr.find(name=faked_hba.name)

        # Execute the code to be tested
        repr_str = repr(hba)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{hba.__class__.__name__}\s+at\s+'
            rf'0x{id(hba):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "initial_partition_status, exp_exc", [
            ('stopped', None),
            ('terminated', None),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', None),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', None),
            ('reservation-error', None),
            ('paused', None),
        ]
    )
    def test_hba_delete(self, initial_partition_status, exp_exc):
        """Test Hba.delete()."""

        # Add a faked HBA to be tested and another one
        faked_hba = self.add_hba1()
        self.add_hba2()

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = initial_partition_status

        hba_mgr = self.partition.hbas

        hba = hba_mgr.find(name=faked_hba.name)

        if exp_exc:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                hba.delete()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the HBA still exists
            hba_mgr.find(name=faked_hba.name)

        else:

            # Execute the code to be tested.
            hba.delete()

            # Check that the HBA no longer exists
            with pytest.raises(NotFound) as exc_info:
                hba_mgr.find(name=faked_hba.name)

    def test_hba_delete_create_same_name(self):
        """Test Hba.delete() followed by Hba.create() with same name."""

        # Add a faked HBA to be tested and another one
        faked_hba = self.add_hba1()
        hba_name = faked_hba.name
        self.add_hba2()

        # Construct the input properties for a third HBA with same name
        part3_props = copy.deepcopy(faked_hba.properties)
        part3_props['description'] = 'Third HBA'

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = 'stopped'  # deletable

        hba_mgr = self.partition.hbas
        hba = hba_mgr.find(name=hba_name)

        # Execute the deletion code to be tested.
        hba.delete()

        # Check that the HBA no longer exists
        with pytest.raises(NotFound):
            hba_mgr.find(name=hba_name)

        # Execute the creation code to be tested.
        hba_mgr.create(part3_props)

        # Check that the HBA exists again under that name
        hba3 = hba_mgr.find(name=hba_name)
        description = hba3.get_property('description')
        assert description == 'Third HBA'

    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New HBA description'},
            {'device-number': 'FEDC',
             'description': 'New HBA description'},
        ]
    )
    def test_hba_update_properties(self, input_props):
        """Test Hba.update_properties()."""

        # Add a faked HBA
        faked_hba = self.add_hba1()

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = 'stopped'  # updatable

        hba_mgr = self.partition.hbas
        hba = hba_mgr.find(name=faked_hba.name)

        hba.pull_full_properties()
        saved_properties = copy.deepcopy(hba.properties)

        # Execute the code to be tested
        hba.update_properties(properties=input_props)

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in hba.properties
            prop_value = hba.properties[prop_name]
            assert prop_value == exp_prop_value

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        hba.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in hba.properties
            prop_value = hba.properties[prop_name]
            assert prop_value == exp_prop_value

    def test_hba_update_name(self):
        """Test Hba.update_properties() with 'name' property."""

        # Add a faked HBA
        faked_hba = self.add_hba1()
        hba_name = faked_hba.name

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = 'stopped'  # updatable

        hba_mgr = self.partition.hbas
        hba = hba_mgr.find(name=hba_name)

        new_hba_name = "new-" + hba_name

        # Execute the code to be tested
        hba.update_properties(properties={'name': new_hba_name})

        # Verify that the resource is no longer found by its old name, using
        # list() (this does not use the name-to-URI cache).
        hbas_list = hba_mgr.list(
            filter_args=dict(name=hba_name))
        assert len(hbas_list) == 0

        # Verify that the resource is no longer found by its old name, using
        # find() (this uses the name-to-URI cache).
        with pytest.raises(NotFound):
            hba_mgr.find(name=hba_name)

        # Verify that the resource object already reflects the update, even
        # though it has not been refreshed yet.
        assert hba.properties['name'] == new_hba_name

        # Refresh the resource object and verify that it still reflects the
        # update.
        hba.pull_full_properties()
        assert hba.properties['name'] == new_hba_name

        # Verify that the resource can be found by its new name, using find()
        new_hba_find = hba_mgr.find(name=new_hba_name)
        assert new_hba_find.properties['name'] == new_hba_name

        # Verify that the resource can be found by its new name, using list()
        new_hbas_list = hba_mgr.list(
            filter_args=dict(name=new_hba_name))
        assert len(new_hbas_list) == 1
        new_hba_list = new_hbas_list[0]
        assert new_hba_list.properties['name'] == new_hba_name

    @pytest.mark.parametrize(
        "initial_partition_status, exp_exc", [
            ('stopped', None),
            ('terminated', None),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', None),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', None),
            ('reservation-error', None),
            ('paused', None),
        ]
    )
    def test_hba_reassign_port(self, initial_partition_status, exp_exc):
        """Test Hba.reassign_port()."""

        # Add a faked HBA to be tested.
        # Its port points to a faked URI.
        faked_hba = self.add_hba1()

        # Add a faked FCP with one port that the HBA will be reassigned to
        faked_adapter = self.faked_cpc.adapters.add({
            'object-id': 'fake-fcp1-oid',
            # object-uri is auto-set based upon object-id
            'parent': self.faked_cpc.uri,
            'class': 'adapter',
            'name': 'fake-fcp1',
            'description': 'FCP #1',
            'status': 'active',
            'type': 'fcp',
            # adapter-family is auto-set based upon type
            'adapter-id': '123',
            'detected-card-type': 'ficon-express-16s',
            'card-location': '1234-5678-J.01',
            'port-count': 1,
            'storage-port-uris': [],
            'state': 'online',
            'configured-capacity': 80,
            'used-capacity': 0,
            'allowed-capacity': 80,
            'maximum-total-capacity': 80,
            'channel-path-id': '1B',
            'physical-channel-status': 'operating',
        })
        adapter = self.cpc.adapters.find(name='fake-fcp1')
        faked_adapter.ports.add({
            'element-id': 'fake-port1-oid',
            # element-uri is auto-set based upon object-id
            'parent': faked_adapter.uri,
            'class': 'storage-port',
            'name': 'fake-port1',
            'description': 'FCP #1 Port 1',
            'index': 0,
            'fabric-id': None,
        })
        port = adapter.ports.find(name='fake-port1')

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = initial_partition_status

        # The HBA object we will perform the test on
        hba = self.partition.hbas.find(name=faked_hba.name)

        # Save the HBA properties for later comparison
        hba.pull_full_properties()
        saved_properties = copy.deepcopy(hba.properties)

        if exp_exc:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                hba.reassign_port(port)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the port of the HBA is unchanged ...
            prop_name = 'adapter-port-uri'
            # ... in the resource object:
            assert hba.properties[prop_name] == saved_properties[prop_name]
            # ... and again when refreshed from the mock state:
            hba.pull_full_properties()
            assert hba.properties[prop_name] == saved_properties[prop_name]

        else:

            # Execute the code to be tested.
            hba.reassign_port(port)

            # Check that the port of the HBA has been set ...
            # ... in the resource object:
            prop_name = 'adapter-port-uri'
            assert hba.properties[prop_name] == port.uri
            # ... and again when refreshed from the mock state:
            hba.pull_full_properties()
            assert hba.properties[prop_name] == port.uri



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`NIC` (Network Interface Card) is a logical entity that provides a
:term:`Partition` with access to external communication networks through a
:term:`Network Adapter`. More specifically, a NIC connects a Partition with a
:term:`Network Port`, or with a :term:`Virtual Switch` which then connects to
the Network Port.

NIC resources are contained in Partition resources.

NICs only exist in :term:`CPCs <CPC>` that are in DPM mode.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._exceptions import ConsistencyError
from ._utils import RC_NIC

__all__ = ['NicManager', 'Nic']


class NicManager(BaseManager):
    """
    Manager providing access to the :term:`NICs <NIC>` in a particular
    :term:`Partition`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Partition` object (in DPM mode):

    * :attr:`~zhmcclient.Partition.nics`

    HMC/SE version requirements:

    * SE version 2.13.1
    """

    def __init__(self, partition):
        # This function should not go into the docs.
        # Parameters:
        #   partition (:class:`~zhmcclient.Partition`):
        #     Partition defining the scope for this manager.

        super().__init__(
            resource_class=Nic,
            class_name=RC_NIC,
            session=partition.manager.session,
            parent=partition,
            base_uri=f'{partition.uri}/nics',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=[],
            list_has_name=False)

    @property
    def partition(self):
        """
        :class:`~zhmcclient.Partition`: :term:`Partition` defining the scope
        for this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the NICs in this Partition.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the corresponding array property for this resource in the
          parent object is used to list the resources, and the provided filter
          arguments are applied.

        HMC/SE version requirements:

        * SE version 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Nic` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        return self._list_with_parent_array(
            self.partition, 'nic-uris', full_properties, filter_args)

    @logged_api_call
    def create(self, properties):
        """
        Create and configure a NIC in this Partition.

        The NIC must be backed by an adapter port (on an OSA, ROCE, or
        Hipersockets adapter).

        The way the backing adapter port is specified in the "properties"
        parameter of this method depends on the adapter type, as follows:

        * For OSA and Hipersockets adapters, the "virtual-switch-uri"
          property is used to specify the URI of the virtual switch that is
          associated with the backing adapter port.

          This virtual switch is a resource that automatically exists as soon
          as the adapter resource exists. Note that these virtual switches do
          not show up in the HMC GUI; but they do show up at the HMC REST API
          and thus also at the zhmcclient API as the
          :class:`~zhmcclient.VirtualSwitch` class.

          The value for the "virtual-switch-uri" property can be determined
          from a given adapter name and port index as shown in the following
          example code (omitting any error handling):

          .. code-block:: python

              partition = ...  # Partition object for the new NIC

              adapter_name = 'OSA #1'  # name of adapter with backing port
              adapter_port_index = 0   # port index of backing port

              adapter = partition.manager.cpc.adapters.find(name=adapter_name)

              vswitches = partition.manager.cpc.virtual_switches.findall(
                  **{'backing-adapter-uri': adapter.uri})

              vswitch = None
              for vs in vswitches:
                  if vs.get_property('port') == adapter_port_index:
                      vswitch = vs
                      break

              properties['virtual-switch-uri'] = vswitch.uri

        * For RoCE adapters, the "network-adapter-port-uri" property is used to
          specify the URI of the backing adapter port, directly.

          The value for the "network-adapter-port-uri" property can be
          determined from a given adapter name and port index as shown in the
          following example code (omitting any error handling):

          .. code-block:: python

              partition = ...  # Partition object for the new NIC

              adapter_name = 'ROCE #1'  # name of adapter with backing port
              adapter_port_index = 0   # port index of backing port

              adapter = partition.manager.cpc.adapters.find(name=adapter_name)

              port = adapter.ports.find(index=adapter_port_index)

              properties['network-adapter-port-uri'] = port.uri

        HMC/SE version requirements:

        * SE version 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Object-access permission to the backing Adapter for the new NIC.
        * Task permission to the "Partition Details" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create NIC' in the :term:`HMC API` book.

        Returns:

          Nic:
            The resource object for the new NIC.
            The object will have its 'element-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.partition.uri + '/nics',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        nic = Nic(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return nic


class Nic(BaseResource):
    """
    Representation of a :term:`NIC`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of a NIC resource, see section
    'Data model - NIC Element Object' in section 'Partition object' in the
    :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.NicManager`).

    HMC/SE version requirements:

    * SE version 2.13.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.NicManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, NicManager), (
            f"Nic init: Expected manager type {NicManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this NIC.

        HMC/SE version requirements:

        * SE version 2.13.1

        Authorization requirements:

        * Object-access permission to the Partition containing this HBA.
        * Task permission to the "Partition Details" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self._uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))

        parent_nic_uris = self.manager.parent.get_properties_local(
            'nic-uris')
        if parent_nic_uris:
            try:
                parent_nic_uris.remove(self._uri)
            except ValueError:
                pass

        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this NIC.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version 2.13.1

        Authorization requirements:

        * Object-access permission to the Partition containing this NIC.
        * Object-access permission to the backing Adapter for this NIC.
        * Task permission to the "Partition Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model - NIC Element Object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, body=properties, resource=self)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    @logged_api_call
    def backing_port(self):
        """
        Return the backing adapter port of the NIC.

        For vswitch-based NICs (e.g. OSA, HiperSocket), the NIC references the
        backing vswitch which references its backing adapter and port index.
        In this case, 'Retrieve ... Properties' operations are performed for
        the vswitch, the adapter and the port.

        For port-based NICs (e.g. RoCE, CNA), the NIC references directly the
        backing port.
        In this case, 'Retrieve ... Properties' operations are performed for
        the port and the adapter.

        HMC/SE version requirements:

        * SE version 2.13.1

        Returns:

            Port: A resource object for the backing adapter port.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        partition = self.manager.parent
        cpc = partition.manager.parent
        client = cpc.manager.client

        # Handle vswitch-based NICs (OSA, HS)
        try:
            vswitch_uri = self.get_property('virtual-switch-uri')
        except KeyError:
            pass
        else:
            vswitch_props = client.session.get(vswitch_uri)
            adapter_uri = vswitch_props['backing-adapter-uri']
            port_index = vswitch_props['port']
            adapter = cpc.adapters.resource_object(adapter_uri)
            port_uris = adapter.get_property('network-port-uris')
            for port_uri in port_uris:
                port = adapter.ports.resource_object(port_uri)
                port_index_ = port.get_property('index')
                if port_index_ == port_index:
                    break
            else:
                raise ConsistencyError(
                    f"HMC inconsistency: NIC {cpc.name}.{partition.name}."
                    f"{self.name} (URI {self.uri}) has a backing adapter "
                    f"{cpc.name}.{adapter.name} (URI {adapter.uri}) that does "
                    f"not have a port with the required port index "
                    f"{port_index}")
            return port

        # Handle port-based NICs (RoCE, CNA)
        try:
            port_uri = self.get_property('network-adapter-port-uri')
        except KeyError:
            pass
        else:
            port_props = client.session.get(port_uri)
            adapter_uri = port_props['parent']
            adapter = cpc.adapters.resource_object(adapter_uri)
            port = adapter.ports.resource_object(port_uri)
            return port

        raise ConsistencyError(
            f"HMC inconsistency: NIC {cpc.name}.{partition.name}.{self.name} "
            f"(URI {self.uri}) has neither 'virtual-switch-uri' nor "
            "'network-adapter-port-uri' properties")



#!/usr/bin/env python
# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that inspects an HMC and dumps it as a HMC definition file.
"""

import sys
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

hmc_yaml_file = 'hmc_definition.yaml'

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Dumping HMC resources as an HMC definition ...")
    hmc_yaml_str = client.to_hmc_yaml()

    print(f"Writing HMC definition to file: {hmc_yaml_file}")
    with open(hmc_yaml_file, 'w') as fp:
        fp.write(hmc_yaml_str)

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for adapter ports (on CPCs in DPM mode).

These tests do not change any existing ports, but create and delete Hipersocket
adapters and modify their ports.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic Port objects (e.g. find_by_name())
PORT_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in Port objects returned by list() without full props
PORT_LIST_PROPS = ['element-uri', 'name', 'description']

# Properties whose values can change between retrievals of Port objects
PORT_VOLATILE_PROPS = []


def test_port_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the ports to test with
        adapter_port_tuples = []
        adapter_list = cpc.adapters.list()
        for adapter in adapter_list:
            port_list = adapter.ports.list()
            for port in port_list:
                adapter_port_tuples.append((adapter, port))
        if not adapter_port_tuples:
            skip_warn(
                f"No adapters with ports on CPC {cpc.name} managed by HMC "
                f"{hd.host}")
        adapter_port_tuples = pick_test_resources(adapter_port_tuples)

        for adapter, port in adapter_port_tuples:
            print(f"Testing on CPC {cpc.name} with port {port.name!r} of "
                  f"adapter {adapter.name!r}")
            runtest_find_list(
                session, adapter.ports, port.name, 'name', 'element-uri',
                PORT_VOLATILE_PROPS, PORT_MINIMAL_PROPS, PORT_LIST_PROPS)


def test_port_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the ports to test with
        adapter_port_tuples = []
        adapter_list = cpc.adapters.list()
        for adapter in adapter_list:
            port_list = adapter.ports.list()
            for port in port_list:
                adapter_port_tuples.append((adapter, port))
        if not adapter_port_tuples:
            skip_warn(
                f"No adapters with ports on CPC {cpc.name} managed by "
                f"HMC {hd.host}")
        adapter_port_tuples = pick_test_resources(adapter_port_tuples)

        for adapter, port in adapter_port_tuples:
            print(f"Testing on CPC {cpc.name} with port {port.name!r} of "
                  f"adapter {adapter.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(port.manager, non_list_prop)


def test_port_update(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test updating the port of a Hipersocket adapter.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        print(f"Testing on CPC {cpc.name}")

        adapter_name = TEST_PREFIX + ' test_adapter_crud adapter1'

        # Ensure a clean starting point for this test
        try:
            adapter = cpc.adapters.find(name=adapter_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test Hipersocket adapter from previous run: "
                f"{adapter_name!r} on CPC {cpc.name}", UserWarning)
            adapter.delete()

        adapter = None
        try:
            # Create a Hipersocket adapter
            adapter_input_props = {
                'name': adapter_name,
            }
            adapter = cpc.adapters.create_hipersocket(adapter_input_props)

            # Pick a port for the test
            port = adapter.ports.list()[0]

            port_name = port.name
            port_name_new = port_name + ' new'

            # Test updating a property of the port

            new_desc = "Updated port description."

            # The code to be tested
            port.update_properties(dict(description=new_desc))

            assert port.properties['description'] == new_desc
            port.pull_full_properties()
            assert port.properties['description'] == new_desc

            # Test that ports cannot be renamed

            with pytest.raises(zhmcclient.HTTPError) as exc_info:

                # The code to be tested
                port.update_properties(dict(name=port_name_new))

            exc = exc_info.value
            assert exc.http_status == 400
            assert exc.reason == 6

        finally:
            # Cleanup
            if adapter:
                adapter.delete()



#!/usr/bin/env python

import sys

if len(sys.argv) != 2:
    print("Error: Incorrect number of command line arguments")
    print(f"Usage: {sys.argv[0]}  version_level")
    print("Where version_level is the number of components in the "
          "version string from 1 to 3, e.g. 3 means Python 2.7.11")

    sys.exit(2)

version_level = sys.argv[1]
if version_level not in ('1', '2', '3'):
    print("Error: Invalid version level; must be 1, 2, or 3")
    sys.exit(2)

version_level = int(version_level)
py_version = '.'.join([str(v) for v in sys.version_info[0:version_level]])
print(py_version)
